program methane_main

  use methane_parameters, only: dp,n_flux_parts, &
    mdiff,mplant,mebuair,mebutot,mprod,mdecay      ,moxi, &
    cdiff,cplant,cebuair,cebutot,cprod,cdecay,cresp,coxi, &
    odiff,oplant,oebuair,oebutot,oprod       ,oresp,ooxi, &
    por, peat_air_coeff, peat_water_coeff, lambda_root, V_R0, K_R, K_CH4, &
    K_O2, get_parameters, root_km, f_exu, exu_pool_turnovertime, Q10, f_methane,&
    Rgas,atm,fm
  use methane_geometry, only: geometry,new_geom,gas_move_wtd
  use methane_output, only: R_output

  implicit none

  ! time loop control:
  integer  :: time_i,n_steps,day_first,day_last
  real(dp) :: time,dt

  ! miscellaneous
  integer :: i

  ! geometry

  ! In principle, we set the model grid to have 'nlayers' background layers,
  ! but we need one extra layer to deal with the moving water table.
  integer :: nlayers
  integer :: nz ! will always be nlayers + 1

  ! naming convention: arrays of size = nz end with P (P = "profile"), also
  ! levelsP although its size is nz+1

  real(dp) :: wtd,wtd_eff,wtd_new ! por,acro_depth
  real(dp), allocatable :: fixed_levelsP(:)
  type(geometry) :: geom

  ! fixed_levelsP: coordinates in [meter] of levels that separate the layers in
  ! the background grid. The moving level will then be inside one of these
  ! layers, at water table level. Or perhaps above fixed_levels(1), if water is
  ! above the peat surface.
  ! geom: see type contents in methane_geometry. Contains:
  !       levelsP, dzP (arrays)
  !       surfaceQ, airQ, waterQ (logical)
  !       s1, s2, a1, a2, w1, w2 (integer)
  ! wtd: water table depth (WTD), positive is below surface, negative
  !      value means the water surface is above peat surface
  ! wtd_eff: effective wtd, returned by subroutine new_geom. If the input wtd
  !          is too close (parameter min_dz, usually 1 cm) to a fixed level,
  !          we don't accept the value, but round to the fixed level, thus
  !          avoiding to ever have a layer less than 1 cm thick.
  ! acro_depth: acrotelm depth, hopefully never below wtd
  ! por: peat porosity, volume fraction of air or water in peat, we use 80%

  ! roots
  real(dp) :: root_mass
  real(dp), allocatable :: rootsP(:)
  ! lambda_root: root depth distribution decay parameter, [m] - now in methane_parameters
  ! rootsP: fraction of roots in each layer, sums to 1
  ! root_mass: root (dry)biomass per m2 of peatland [kg]
  ! root_km: root-ending area per root (dry)biomass [m2/kg]

  ! temperature
  real(dp) :: T_surface,T_deep
  real(dp), allocatable :: tempP(:)
  ! real(dp), allocatable :: k_thermalP(:),vhcP(:)
  ! tempP:      temperature (Kelvin) in layers
  ! k_thermalP: thermal conductivity in layers [W m-1 K-1]
  ! vhcP:       volumetric heat capacity of layers [J m-3 K-1]

  ! methane, CO2 and oxygen
  real(dp), parameter :: methane_mole_fraction = 1.8e-6_dp ! [mol mol(dry air)-1]
  real(dp), parameter :: co2_mole_fraction = 400.e-6_dp ! [mol mol(dry air)-1]
  real(dp), parameter :: oxygen_mole_fraction = 0.2095_dp ! [mol mol(dry air)-1]
  ! xP: amount of x in each layer [mol], really [mol] not [mol m-3]
  real(dp), allocatable :: methaneP(:) ! [mol]
  real(dp), allocatable :: co2P(:) ! [mol]
  real(dp), allocatable :: oxygenP(:) ! [mol]

  ! x_Ca: concentration of x in ambient air [mol m-3]
  real(dp) :: methane_Ca,co2_Ca,oxygen_Ca

  ! temporary
  real(dp), allocatable :: methane_newP(:)
  real(dp), allocatable :: co2_newP(:)
  real(dp), allocatable :: oxygen_newP(:)
  real(dp), allocatable :: alphaP(:),CairP(:)

  ! soil carbon
  real(dp) :: exu_rate,exu_pool, &
              exu_pool_turnoverrate
  ! exu_rate:              root exudation rate [mol C s-1]
  ! exu_pool:              underground exudates carbon pool [mol]
  !   To be more "realistic", we might have a carbon bool in each layer.
  !   But now we opt for simplicity, and there is only one pool.
  ! f_exu:                 fraction of NPP carbon, that goes to root exudates
  ! f_methane:             fraction on root exudates (underground carbon) that
  !   produces methane. rest goes to CO2.
  ! exu_pool_turnovertime: turnover time for the soil carbon pool.
  ! exu_pool_turnoverrate: turnover rate = 1/(turnover time)
  !   d(exu_pool)/dt = -turnoverrate*exu_pool + (input rate)


  ! input
  integer, parameter :: n_data = 730, cols_data = 7
  ! when input data changes, need to adjust this by hand
  real(dp), dimension(cols_data,n_data) :: input_data
  integer,  dimension(n_data) :: day_data
  real(dp), dimension(n_data) :: temp_data,wtd_data,cato_data,lai_data, &
                                 npp_data,root_data
  real(dp), dimension(7,n_data) :: ztemp_data ! soil temperature in 20 layers

  ! output
  real(dp) :: gases_out(n_flux_parts),gases_out_day(n_flux_parts), &
              gases_out_tot(n_flux_parts),gases_out_rates(n_flux_parts)

  ! *** methane transport ***
  !  1 = methane diffusion
  !  2 = methane plant transport
  !  3 = methane ebullition into air
  !  4 = methane ebullition total (when WTD is below peat surface, ebullition is
  !      released in the lowest air layer inside peat)
! *** methane production & loss ***
  !  5 = production of methane from exu (also not a flux)
  !  6 = production of methane from peat decay in catotelm (below WT)
  !  7 = oxidation of methane (not a flux out, just a sink)
! *** CO2 transport ***
  !  8 = CO2 diffusion
  !  9 = CO2 plant transport
  ! 10 = CO2 ebullition into air (CO2 ebullition is assumed to occur with methane)
  ! 11 = CO2 ebullition total (when WTD is below peat surface, ebullition is
  !      released in the lowest air layer inside peat)
! *** CO2 production ***
  ! 12 = production of CO2 from exu (also not a flux)
  ! 13 = production of CO2 from peat decay (also not a flux)
  ! 14 = production of CO2 from methane oxidation (not a flux out)
  ! 15 = production of CO2 from respiration (also not a flux)
! *** Oxygen transport ***
  ! 16 = oxygen diffusion
  ! 17 = oxygen plant transport
  ! 18 = oxygen ebullition into air (oxygen ebullition is assumed to occur with methane)
  ! 19 = oxygen ebullition total (when WTD is below peat surface, ebullition is
  !      released in the lowest air layer inside peat)
! *** Oxygen consumption ***
  ! 20 = oxygen consumed in oxidation of root exudates
  ! 21 = oxygen consumed in respiration
  ! 22 = oxygen consumed in oxidation of methane

!  real(dp) :: emission_tot

  ! mass control
  real(dp) :: methane_time1,methane_time2, &
              methane_wtd_flux,methane_wtd_flux_tot, &
              methane_produced,methane_escaped, &
              co2_time1,co2_time2, &
              co2_wtd_flux,co2_wtd_flux_tot, &
              co2_produced,co2_escaped, &
! O2
              oxygen_time1,oxygen_time2, &
              oxygen_wtd_flux,oxygen_wtd_flux_tot, &
              o2_consumed,o2_input,o2_escaped

  integer :: i_wtd   ! this is the layer containing wtd
  
  ! Get parameters from namelist
  call get_parameters

  nlayers = 7      ! number of layers without the extra layer for WTD
  nz = nlayers + 1 ! number of layers including the extra moving layer

  ! For now, it would have been simpler to have defined these as constant
  ! length arrays, but I aim for full generality for later.
  allocate(fixed_levelsP(nz),geom%levelsP(nz+1),geom%dzP(nz))
  allocate(methaneP(nz),methane_newP(nz))
  allocate(co2P(nz),co2_newP(nz))
  allocate(oxygenP(nz),oxygen_newP(nz))
  allocate(rootsP(nz))
  allocate(alphaP(nz), CairP(nz))
  allocate(tempP(nz))

  ! initialize height levels
  do i = 1,nz
    fixed_levelsP(i) = (i-1) * 0.1_dp
  end do

fixed_levelsP(nz) = 4.-0.1*(nz-2.)
  
  ! read input data
  open(11,file='input.txt',action='READ')
  read(11,*) input_data
  close(11)

  open(12, file='ztemp_data.txt', action='read')
  read(12, *) ztemp_data
  close(12)

  day_data  = floor(input_data(1,:)) ! day number
  temp_data = input_data(2,:) ! temperature
  wtd_data  = input_data(3,:) ! wtd depth
  cato_data = input_data(4,:) ! acrotelm-catotelm boundary depth
  lai_data  = input_data(5,:) ! vascular plant LAI
  npp_data  = input_data(6,:) ! vascular plant NPP [mol m-2 s-1]
  root_data = input_data(7,:) ! vascular plant root dry biomass [kg m-2]

  day_first = day_data(1)
  day_last  = day_data(size(day_data)) ! takes the last element

  ! initialize geometry

  wtd_new = wtd_data(1)
  !wtd_new = 1.0 ! debug
  call new_geom(geom,wtd_eff,nz,fixed_levelsP,wtd_new)
  wtd = wtd_eff

  write(*,*) 'levels:'
  write(*,'(11(9F7.3))') geom%levelsP
  write(*,*) 'wtd:', wtd

  rootsP(1:nz) = &
       calculate_root_fraction_alternative(lambda_root,nz,geom%levelsP,fixed_levelsP(1)) ! Shifted Gaussian

  write(*,*) 'roots:'
  write(*,'(10(6F6.3))') rootsP

  T_surface = 273.15 + 10
  T_deep = 273.15 + 10
  tempP = 273.15 + 10

  methane_Ca = methane_mole_fraction*atm/(Rgas*T_surface) ! [mol m-3]
  co2_Ca = co2_mole_fraction*atm/(Rgas*T_surface) ! [mol m-3]
  oxygen_Ca = oxygen_mole_fraction*atm/(Rgas*T_surface) ! [mol m-3]
  ! not used
  ! CwaterP = 0
  CairP = 0

  exu_pool_turnoverrate = 1.0/exu_pool_turnovertime

  co2P = 0
  oxygenP  = 0
  exu_pool = 0

  open(13, file='methaneP.txt', action='read')
  read(13, *) methaneP
  methane_newp = methanep
  close(13)

  ! time loop preparation
  time = day_first * 24*3600
  dt = 24*3600 ! 1 day in seconds

  gases_out = 0 ; gases_out_day = 0 ; gases_out_tot = 0

  methane_time1 = sum(methaneP)
  co2_time1 = sum(co2P)
  oxygen_time1 = sum(oxygenP)

  methane_wtd_flux_tot = 0
  co2_wtd_flux_tot = 0
  oxygen_wtd_flux_tot = 0

  gases_out_rates = gases_out / dt
  call R_output(time,dt,nz,geom,por,methaneP,co2P,oxygenP,gases_out_rates,exu_pool)

  n_steps = day_last - day_first + 1
  do time_i = 1, n_steps ! main time loop ********************************

    time = time + dt
    wtd_new = wtd_data(time_i)
    call new_geom(geom,wtd_eff,nz,fixed_levelsP,wtd_new)

    ! If WTD has moved, adjust gas (and root) content in gridboxes
    if (wtd /= wtd_eff) then
      rootsP(1:nz) = &
        calculate_root_fraction_alternative(lambda_root,nz,geom%levelsP,fixed_levelsP(1))

      ! should verify the wtd_flux
      alphaP = methane_solubility(tempP) ! [m(ch4)3 m(water)-3]
      call gas_move_wtd(methane_newP,methane_wtd_flux,nz,wtd,wtd_eff,methaneP,alphaP,fixed_levelsP)
      methaneP = methane_newP

      alphaP = co2_solubility(tempP) ! [m(co2)3 m(water)-3]
      call gas_move_wtd(co2_newP,co2_wtd_flux,nz,wtd,wtd_eff,co2P,alphaP,fixed_levelsP)
      co2P = co2_newP

      alphaP = oxygen_solubility(tempP) ! [m(o2)3 m(water)-3]
      call gas_move_wtd(oxygen_newP,oxygen_wtd_flux,nz,wtd,wtd_eff,oxygenP,alphaP,fixed_levelsP)
      oxygenP = oxygen_newP

      wtd = wtd_eff
      methane_wtd_flux_tot = methane_wtd_flux_tot + methane_wtd_flux
      co2_wtd_flux_tot = co2_wtd_flux_tot + co2_wtd_flux
      oxygen_wtd_flux_tot  = oxygen_wtd_flux_tot  + oxygen_wtd_flux
    end if

    ! read in (in the future) surface temperature

    ! Update soil temperature. Data from i_wtd layer goes to both
    i_wtd = max(nint(wtd_new/0.1+0.5),1)
    if (i_wtd .gt. nlayers) then
       i_wtd = nlayers
    end if

    tempP(:i_wtd) = ztemp_data(:i_wtd,time_i) + 273.15
    tempP(i_wtd+1:) = ztemp_data(i_wtd:,time_i) + 273.15
    T_surface = 273.15 + 10

    ! heat transfer goes here...

    exu_rate  = f_exu * npp_data(time_i)
    root_mass = root_data(time_i)

    call bigstep(methaneP,co2P,oxygenP,exu_pool,gases_out, &
      nz,geom,por,dt,tempP,methane_Ca,co2_Ca,oxygen_Ca, &
      exu_rate,exu_pool_turnoverrate,f_methane, &
      rootsP,root_km,root_mass)

    gases_out_rates = gases_out / dt
    call R_output(time,dt,nz,geom,por,methaneP,co2P,oxygenP,gases_out_rates,exu_pool)

    gases_out_tot = gases_out_tot + gases_out
  end do ! end of main time loop *********************************************

  methane_time2 = sum(methaneP)
  co2_time2 = sum(co2P)
  oxygen_time2 = sum(oxygenP)

  write(*,*) ; write(*,*) 'Final results'
  write(*,*) 'levelsP:' ; write(*,'(11(11F9.3))') geom%levelsP
  write(*,*)
  write(*,*) 'methane_Ca:', methane_Ca
  write(*,*) 'methane:' ; write(*,'(11(11F12.6))') methaneP
  write(*,*)
  write(*,*) 'co2_Ca:', co2_Ca
  write(*,*) 'CO2:' ; write(*,'(11(11F12.6))') co2P
  write(*,*)
  write(*,*) 'oxygen_Ca:', oxygen_Ca
  write(*,*) 'oxygen:' ; write(*,'(11(11F12.6))') oxygenP
  write(*,*)
  write(*,*) 'exu_pool, f_methane:', exu_pool, f_methane
  write(*,*)
  write(*,*) 'Methane results:'
  write(*,*) 'column total, t_start: ',methane_time1
  write(*,*) 'column total, t_end:   ',methane_time2
  methane_produced = gases_out_tot(mprod) + gases_out_tot(mdecay)
  methane_escaped = gases_out_tot(mdiff) + gases_out_tot(mplant) &
       + gases_out_tot(mebuair)
  write(*,*) 'produced: ',methane_produced
  write(*,*) '    produced (roots):              ',gases_out_tot(mprod)
  write(*,*) '    produced (peat decay):         ',gases_out_tot(mdecay)
  write(*,*) 'consumed: ',gases_out_tot(moxi)
  write(*,*) '    oxidized:                      ',gases_out_tot(moxi)
  write(*,*) 'escaped:  ',methane_escaped
  write(*,*) '    escaped due to diffusion:      ',gases_out_tot(mdiff)
  write(*,*) '    escaped due to plant transport:',gases_out_tot(mplant)
  write(*,*) '    escaped due to ebullition:     ',gases_out_tot(mebuair)
  write(*,*) '    (total ebullition was:         ',gases_out_tot(mebutot),')'
  write(*,*) 'change due to moving WTD: ',methane_wtd_flux_tot
  write(*,*) 'balance:  ', -(methane_time2 - methane_time1) &
       + methane_produced + gases_out_tot(moxi) + methane_escaped &
       - methane_wtd_flux_tot, ' mol'
  write(*,*) '(Something like x.xE-014 is acceptable)'
  write(*,*)
  write(*,*) 'CO2 results:'
  write(*,*) 'column total, t_start: ',co2_time1
  write(*,*) 'column total, t_end:   ',co2_time2
  co2_produced = gases_out_tot(cprod) + gases_out_tot(cdecay) &
       + gases_out_tot(cresp) + gases_out_tot(coxi)
  co2_escaped = gases_out_tot(cdiff) + gases_out_tot(cplant) &
       + gases_out_tot(cebuair)
  write(*,*) 'produced: ',co2_produced
  write(*,*) '    produced (roots):              ',gases_out_tot(cprod)
  write(*,*) '    produced (peat decay):         ',gases_out_tot(cdecay)
  write(*,*) '    produced (respiration):        ',gases_out_tot(cresp)
  write(*,*) '    produced (CH4 oxidation):      ',gases_out_tot(coxi)
  write(*,*) 'escaped:  ',co2_escaped
  write(*,*) '    escaped due to diffusion:      ',gases_out_tot(cdiff)
  write(*,*) '    escaped due to plant transport:',gases_out_tot(cplant)
  write(*,*) '    escaped due to ebullition:     ',gases_out_tot(cebuair)
  write(*,*) '    (total ebullition was:         ',gases_out_tot(cebutot),')'
  write(*,*) 'change due to moving WTD: ',co2_wtd_flux_tot
  write(*,*) 'balance:  ', -(co2_time2 - co2_time1) &
       + co2_produced + co2_escaped &
       - co2_wtd_flux_tot, 'mol'
  write(*,*)
  write(*,*) 'Oxygen results:'
  write(*,*) 'column total, t_start: ',oxygen_time1
  write(*,*) 'column total, t_end:   ',oxygen_time2
  o2_consumed = gases_out_tot(oprod) + gases_out_tot(oresp) + gases_out_tot(ooxi)
  o2_input = gases_out_tot(odiff) + gases_out_tot(oplant)
  o2_escaped = gases_out_tot(oebuair)
  write(*,*) 'consumed:        ',o2_consumed
  write(*,*) '    consumed in oxidation of exu: ',gases_out_tot(oprod)
  write(*,*) '    consumed in respiration:      ',gases_out_tot(oresp)
  write(*,*) '    consumed in methane oxidation:',gases_out_tot(ooxi)
  write(*,*) 'transport input: ',o2_input
  write(*,*) '    diffusion input:              ',gases_out_tot(odiff)
  write(*,*) '    plant transport input:        ',gases_out_tot(oplant)
  write(*,*) 'escaped:         ',o2_escaped
  write(*,*) '    escaped due to ebullition:    ',gases_out_tot(oebuair)
  write(*,*) '    (total ebullition was:        ',gases_out_tot(oebutot),')'
  write(*,*) 'change due to moving WTD: ',oxygen_wtd_flux_tot
  write(*,*) 'balance:         ', -(oxygen_time2 - oxygen_time1) &
       + o2_consumed + o2_input + o2_escaped &
       - oxygen_wtd_flux_tot, 'mol'
contains

  subroutine bigstep(methaneP,co2P,oxygenP,exu_pool,gases_out_bigstep, &
    nz,geom,por,big_dt,tempP,methane_Ca,co2_Ca,oxygen_Ca, &
    exu_rate,exu_pool_turnoverrate,f_methane,rootsP,root_km,root_mass)
    integer, intent(in)        :: nz
    real(dp), intent(inout)    :: methaneP(nz),co2P(nz),oxygenP(nz),exu_pool
    real(dp), intent(out)      :: gases_out_bigstep(n_flux_parts)
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: por,big_dt,tempP(nz),methane_Ca,co2_Ca, &
                                  oxygen_Ca,exu_rate,exu_pool_turnoverrate, &
                                  f_methane,rootsP(nz),root_km,root_mass
    real(dp)                   :: met_d1(nz),met_d2(nz),met_d3(nz),met_d4(nz), &
                                  met_d(nz)
    real(dp)                   :: co2_d1(nz),co2_d2(nz),co2_d3(nz),co2_d4(nz), &
                                  co2_d(nz)
    real(dp)                   :: oxy_d1(nz),oxy_d2(nz),oxy_d3(nz), &
                                  oxy_d4(nz),oxy_d(nz)
    real(dp)                   :: exu_pool_d1,exu_pool_d2,&
                                  exu_pool_d3,exu_pool_d4,exu_pool_d
    real(dp)                   :: f_out1(n_flux_parts), &
                                  f_out2(n_flux_parts), &
                                  f_out3(n_flux_parts), &
                                  f_out4(n_flux_parts)
    real(dp)                   :: met_DaP(nz),met_DwP(nz),met_alphaP(nz), &
                                  met_DcondP(nz),met_root_condP(nz)
    real(dp)                   :: co2_DaP(nz),co2_DwP(nz),co2_alphaP(nz), &
                                  co2_DcondP(nz),co2_root_condP(nz)
    real(dp)                   :: oxy_DaP(nz),oxy_DwP(nz),oxy_alphaP(nz), &
                                  oxy_DcondP(nz),oxy_root_condP(nz)
    real(dp)                   :: max_dt_met,max_dt_co2,max_dt_oxy
    real(dp)                   :: max_diffusion_dt,max_exu_dt,max_respi_dt, &
                                  max_oxid_dt,max_ebu_dt,max_dt,dt
    real(dp)                   :: gases_out(n_flux_parts)
    real(dp)                   :: invpordzP(nz)
    
    integer                    :: n_small_steps, i

    met_DaP    = methane_D_air(tempP)
    met_DwP    = methane_D_water(tempP)
    met_alphaP = methane_solubility(tempP)
    met_DcondP = diffusion_conductances(nz,geom,met_DaP,met_DwP,met_alphaP)
    max_dt_met = find_max_diffusion_dt(nz,geom,met_DcondP)

    co2_DaP    = co2_D_air(tempP)
    co2_DwP    = co2_D_water(tempP)
    co2_alphaP = co2_solubility(tempP)
    co2_DcondP = diffusion_conductances(nz,geom,co2_DaP,co2_DwP,co2_alphaP)
    max_dt_co2 = find_max_diffusion_dt(nz,geom,co2_DcondP)

    oxy_DaP    = oxygen_D_air(tempP)
    oxy_DwP    = oxygen_D_water(tempP)
    oxy_alphaP = oxygen_solubility(tempP)
    oxy_DcondP = diffusion_conductances(nz,geom,oxy_DaP,oxy_DwP,oxy_alphaP)
    max_dt_oxy = find_max_diffusion_dt(nz,geom,oxy_DcondP)

    met_root_condP = root_conductances(nz,geom,met_DaP,rootsP,root_km,root_mass)
    co2_root_condP = root_conductances(nz,geom,co2_DaP,rootsP,root_km,root_mass)
    oxy_root_condP = root_conductances(nz,geom,oxy_DaP,rootsP,root_km,root_mass)

    max_diffusion_dt = min(max_dt_met,max_dt_co2,max_dt_oxy)
    max_exu_dt = find_max_exudation_dt(exu_pool_turnoverrate)
    max_respi_dt = find_max_respiration_dt(nz,geom,oxygenP,tempP,oxy_alphaP,por)
    max_oxid_dt = find_max_oxidation_dt(methaneP,met_alphaP,oxygenP, &
         oxy_alphaP,tempP,nz,geom,por)

    max_ebu_dt = find_max_ebullition_dt()
    max_dt = min(max_diffusion_dt,max_exu_dt,max_respi_dt,max_oxid_dt,max_ebu_dt)

    if (time_i .lt. 3) max_dt = max_dt/30.
    ! Stabilizes beginning so that we save time
    
    n_small_steps = ceiling(big_dt / max_dt)
    dt = big_dt / n_small_steps

    gases_out_bigstep = 0

    invpordzP = 0
    invpordzP(1:geom%a2) = 1./(por*geom%dzP(1:geom%a2))
    invpordzP(geom%w1:geom%w2) = 1./(por*geom%dzP(geom%w1:geom%w2))
    
    do i = 1, n_small_steps

      ! Runge-Kutta-4
      call derivative(methaneP, &
        co2P, &
        oxygenP, &
        exu_pool, &
        met_d1,co2_d1,oxy_d1,exu_pool_d1,f_out1, &
        nz,geom,por,methane_Ca,co2_Ca,oxygen_Ca, &
        exu_rate,exu_pool_turnoverrate,f_methane,tempP, &
        met_DcondP,co2_DcondP,oxy_DcondP,met_alphaP,co2_alphaP,oxy_alphaP, &
        rootsP,met_root_condP,co2_root_condP,oxy_root_condP,invpordzP)

      call derivative(methaneP + 0.5*dt*met_d1, &
        co2P + 0.5*dt*co2_d1, &
        oxygenP + 0.5*dt*oxy_d1, &
        exu_pool + 0.5*dt*exu_pool_d1, &
        met_d2,co2_d2,oxy_d2,exu_pool_d2,f_out2, &
        nz,geom,por,methane_Ca,co2_Ca,oxygen_Ca, &
        exu_rate,exu_pool_turnoverrate,f_methane,tempP, &
        met_DcondP,co2_DcondP,oxy_DcondP,met_alphaP,co2_alphaP,oxy_alphaP, &
        rootsP,met_root_condP,co2_root_condP,oxy_root_condP,invpordzP)

      call derivative(methaneP + 0.5*dt*met_d2, &
        co2P + 0.5*dt*co2_d2, &
        oxygenP + 0.5*dt*oxy_d2, &
        exu_pool + 0.5*dt*exu_pool_d2, &
        met_d3,co2_d3,oxy_d3,exu_pool_d3,f_out3, &
        nz,geom,por,methane_Ca,co2_Ca,oxygen_Ca, &
        exu_rate,exu_pool_turnoverrate,f_methane,tempP, &
        met_DcondP,co2_DcondP,oxy_DcondP,met_alphaP,co2_alphaP,oxy_alphaP, &
        rootsP,met_root_condP,co2_root_condP,oxy_root_condP,invpordzP)

      call derivative(methaneP + dt*met_d3, &
        co2P + dt*co2_d3, &
        oxygenP + dt*oxy_d3, &
        exu_pool + dt*exu_pool_d3, &
        met_d4,co2_d4,oxy_d4,exu_pool_d4,f_out4, &
        nz,geom,por,methane_Ca,co2_Ca,oxygen_Ca, &
        exu_rate,exu_pool_turnoverrate,f_methane,tempP, &
        met_DcondP,co2_DcondP,oxy_DcondP,met_alphaP,co2_alphaP,oxy_alphaP, &
        rootsP,met_root_condP,co2_root_condP,oxy_root_condP,invpordzP)

      met_d = (met_d1 + 2*met_d2 + 2*met_d3 + met_d4)/6
      co2_d = (co2_d1 + 2*co2_d2 + 2*co2_d3 + co2_d4)/6
      oxy_d = (oxy_d1 + 2*oxy_d2 + 2*oxy_d3 + oxy_d4)/6
      exu_pool_d = (exu_pool_d1 + 2*exu_pool_d2 + 2*exu_pool_d3 + exu_pool_d4)/6
      gases_out = (f_out1 + 2*f_out2 + 2*f_out3 + f_out4)/6

      methaneP = methaneP + dt * met_d
      co2P = co2P + dt * co2_d
      oxygenP  = oxygenP  + dt * oxy_d
      exu_pool = exu_pool + dt * exu_pool_d
      gases_out_bigstep = gases_out_bigstep + dt * gases_out
    end do
  end subroutine bigstep

  ! time derivative of all state variables (arrays), also outputs fluxes
  subroutine derivative(methaneP,co2P,oxygenP,exu_pool, &
    methane_d,co2_d,oxygen_d,exu_pool_d,f_out, &
    nz,geom,por,methane_Ca,co2_Ca,oxygen_Ca, &
    exu_rate,exu_pool_turnoverrate,f_methane,tempP, &
    met_DcondP,co2_DcondP,oxy_DcondP,met_alphaP,co2_alphaP,oxy_alphaP, &
    rootsP,met_root_condP,co2_root_condP,oxy_root_condP, invpordzp)
    real(dp), intent(out)      :: methane_d(nz),co2_d(nz),oxygen_d(nz), &
                                  exu_pool_d,f_out(n_flux_parts)
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: methaneP(nz),co2P(nz),oxygenP(nz),exu_pool, &
                                  por,methane_Ca,co2_Ca,oxygen_Ca, &
                                  exu_rate,exu_pool_turnoverrate,f_methane,tempP(nz), &
                                  met_DcondP(nz),co2_DcondP(nz),oxy_DcondP(nz), &
                                  met_alphaP(nz),co2_alphaP(nz),oxy_alphaP(nz), &
                                  rootsP(nz),met_root_condP(nz),co2_root_condP(nz), &
                                  oxy_root_condP(nz)
    real(dp)                   :: met_diffusionP(nz),co2_diffusionP(nz), &
                                  oxy_diffusionP(nz), &
                                  met_flux,co2_flux,oxy_flux, &
                                  met_prodP(nz),co2_prodP(nz),oxy_prodP(nz), &
                                  met_exu,co2_exu,oxy_exu, &
                                  met_decayP(nz),met_cato_decay, &
                                  co2_decayP(nz),co2_cato_decay, &
                                  exu_decayrate, &
                                  met_transportP(nz),co2_transportP(nz),oxy_transportP(nz), &
                                  oxy_respiP(nz),oxy_respi_tot, &
                                  co2_respiP(nz),co2_respi_tot, &
                                  met_oxidP(nz),met_oxid_tot, &
                                  co2_oxidP(nz),co2_oxid_tot, &
                                  oxy_oxidP(nz),oxy_oxid_tot, &
                                  mebu_rateP(nz),mebu_tot,mebu_out, &
                                  cebu_rateP(nz),cebu_tot,cebu_out, &
                                  oebu_rateP(nz),oebu_tot,oebu_out, &
                                  invpordzP(nz)
    
    
    f_out = 0

    co2_oxidp = 0 ! Let's see if this helps with valgrinding
  
    ! Diffusion
    ! Methane diffusion
    call diffusion(met_diffusionP,met_flux,methaneP,nz,geom,methane_Ca, &
      met_DcondP,met_alphaP,invpordzP)
    f_out(mdiff) = met_flux
    ! CO2 diffusion
    call diffusion(co2_diffusionP,co2_flux,co2P,nz,geom,co2_Ca, &
      co2_DcondP,co2_alphaP,invpordzP)
    f_out(cdiff) = co2_flux    
    ! Oxygen diffusion
    call diffusion(oxy_diffusionP,oxy_flux,oxygenP,nz,geom,oxygen_Ca, &
      oxy_DcondP,oxy_alphaP,invpordzP)   
    f_out(odiff) = oxy_flux

    ! Methane and CO2 generation & oxygen consumption from root exudates pool
    exu_decayrate = exu_pool_turnoverrate * exu_pool
    exu_pool_d = -exu_decayrate + exu_rate
    call methane_from_exu(met_prodP,met_exu,co2_prodP,co2_exu,oxy_prodP,oxy_exu, &
         nz,geom,rootsP,oxygenP,exu_decayrate,f_methane)
    f_out(mprod) = met_exu
    f_out(cprod) = co2_exu
    f_out(oprod) = oxy_exu

    ! Methane & CO2 from peat decay (catotelm = below WT)
    call methane_from_peat(met_decayP,met_cato_decay,co2_decayP,co2_cato_decay, &
         nz,geom,tempP,fm)
    f_out(mdecay) = met_cato_decay
    f_out(cdecay) = co2_cato_decay

    ! Root aerenchyma transport
    ! Methane transport
    call root_transport(met_transportP,met_flux,methaneP,nz,geom, &
      met_root_condP,met_alphaP,methane_Ca,por)
    f_out(mplant) = met_flux
    ! CO2 transport
    call root_transport(co2_transportP,co2_flux,co2P,nz,geom, &
      co2_root_condP,co2_alphaP,co2_Ca,por)
    f_out(cplant) = co2_flux
    ! Oxygen transport
    call root_transport(oxy_transportP,oxy_flux,oxygenP,nz,geom, &
      oxy_root_condP,oxy_alphaP,oxygen_Ca,por)
    f_out(oplant) = oxy_flux

    ! Oxygen consumed & CO2 produced in non-methane-related respiration
    call oxygen_respiration(oxy_respiP,oxy_respi_tot,co2_respiP,co2_respi_tot, &
         nz,geom,oxygenP,tempP,oxy_alphaP,por)
    f_out(oresp) = oxy_respi_tot
    f_out(cresp) = co2_respi_tot

    ! Methane oxidation
    call methane_oxidation(met_oxidP,met_oxid_tot, &
         co2_oxidP,co2_oxid_tot,oxy_oxidP,oxy_oxid_tot, &
         methaneP,met_alphaP,oxygenP,oxy_alphaP,tempP,nz,geom,por)
    f_out(moxi) = met_oxid_tot
    f_out(coxi) = co2_oxid_tot
    f_out(ooxi) = oxy_oxid_tot

    ! Methane, CO2 & oxygen ebullition
    call ebullition_rate(mebu_rateP,mebu_tot,mebu_out,methaneP,met_alphaP, &
         cebu_rateP,cebu_tot,cebu_out,co2P,co2_alphaP, &
         oebu_rateP,oebu_tot,oebu_out,oxygenP,oxy_alphaP, &
         nz,geom,por)
    f_out(mebuair) = mebu_out
    f_out(mebutot) = mebu_tot
    f_out(cebuair) = cebu_out
    f_out(cebutot) = cebu_tot
    f_out(oebuair) = oebu_out
    f_out(oebutot) = oebu_tot

    methane_d = met_diffusionP + met_transportP + mebu_rateP &
                + met_prodP + met_decayP              + met_oxidP 
    co2_d     = co2_diffusionP + co2_transportP + cebu_rateP &
                + co2_prodP + co2_decayP + co2_respiP + co2_oxidP
    oxygen_d  = oxy_diffusionP + oxy_transportP + oebu_rateP &
                + oxy_prodP              + oxy_respiP + oxy_oxidP
  end subroutine derivative

  pure function find_max_ebullition_dt() result(max_dt)
    use methane_parameters, only: ebu_hl,time_scale
    real(dp) :: max_dt
    max_dt = time_scale * ebu_hl
  end function find_max_ebullition_dt

  pure subroutine ebullition_rate(mrateP,mebu_tot,mebu_out,methaneP,met_alphaP, &
       crateP,cebu_tot,cebu_out,co2P,co2_alphaP, &
       orateP,oebu_tot,oebu_out,oxygenP,oxy_alphaP, &
       nz,geom,por)

    use methane_parameters, only: ebu_hl,Rgas,atm,ssr

    real(dp), intent(out)      :: mrateP(nz),mebu_tot,mebu_out, &
         crateP(nz),cebu_tot,cebu_out,orateP(nz),oebu_tot,oebu_out         
    real(dp), intent(in)       :: methaneP(nz),met_alphaP(nz), &
         co2P(nz),co2_alphaP(nz),oxygenP(nz),oxy_alphaP(nz),por
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp)                   :: z_mid(nz)
    real(dp)                   :: ppi(4),dppi(4) ! 4 gases: 1=N2, 2=CH4, 3=CO2, 4=O2
    real(dp)                   :: wtd,z,pressure,fact
    real(dp)                   :: met_Cw,co2_Cw,oxy_Cw
    integer                    :: i,n

    real(dp), parameter        :: dist_atm = 10.3_dp ! 1 atm water height [m]

    real(dp) :: k
    k = log(2.0_dp)/ebu_hl

    ! if levelsP(1) < wtd, 
    ! then bubbles will transport gas from gas-saturated water layers 
    ! to the deepest air layer. 
    !
    ! if wtd <= levelsP(1), 
    ! then the transport is directly to atmosphere.
    !
    ! rateP:   The rate at which ebullition removes gas at each
    !          layer, [mol s-1].
    ! ebu_tot: The total rate ebullition is going on, even though it may not be
    !          transporting gas to atmosphere, but to the deepest air layer.
    ! ebu_out: The rate of ebullition releasing gas to atmosphere.

    ! Bubble formation occurs if the sum of the partial pressures 
    ! of all gases (N2, CH4, CO2, O2) in the water exceeds the sum 
    ! of the hydrostatic and the atmospheric pressures. N2 in water 
    ! is assumed to be in equilibrium with the atmospheric content,
    ! i.e. p(N2) = 0.78*atmospheric pressure.
    ! The temperature profile is not considered for p(N2).
    ! Argon could be added, e.g. by setting p(N2) = 0.79*atmospheric pressure.

    mrateP = 0 ; mebu_tot = 0 ; mebu_out = 0
    crateP = 0 ; cebu_tot = 0 ; cebu_out = 0
    orateP = 0 ; oebu_tot = 0 ; oebu_out = 0

    ppi(1) = 0.78 * atm ! partial pressure of N2 in all layers

    z_mid = (geom%levelsP(1:nz) + geom%levelsP(2:nz+1))/2 ! midpoint depths

    if(geom%surfaceQ) then
      wtd = geom%levelsP(1)
      do i = geom%s1,geom%s2
        z = z_mid(i) - wtd
        pressure = atm * (1 + z/dist_atm)
        met_Cw = methaneP(i) / geom%dzP(i) ! concentration in water
        co2_Cw = co2P(i) / geom%dzP(i)
        oxy_Cw = oxygenP(i) / geom%dzP(i)
        ppi(2) = met_Cw * Rgas * tempP(i) / met_alphaP(i) ! CH4 partial pressure
        ppi(3) = co2_Cw * Rgas * tempP(i) / co2_alphaP(i) ! CO2
        ppi(4) = oxy_Cw * Rgas * tempP(i) / oxy_alphaP(i) ! O2
        if (sum(ppi) > ssr * pressure) then
           fact = (sum(ppi) - pressure) / sum(ppi)
           do n = 1,4
              dppi(n) = fact * ppi(n) ! change in each gas [Pa]
           end do
           mrateP(i) = -k * dppi(2) / (Rgas * tempP(i)) * geom%dzP(i)
           crateP(i) = -k * dppi(3) / (Rgas * tempP(i)) * geom%dzP(i)
           orateP(i) = -k * dppi(4) / (Rgas * tempP(i)) * geom%dzP(i)
        end if
      end do
      mebu_out = sum(mrateP) ; mebu_tot = mebu_out
      cebu_out = sum(crateP) ; cebu_tot = cebu_out
      oebu_out = sum(orateP) ; oebu_tot = oebu_out
    end if

    if(geom%waterQ) then
      wtd = geom%levelsP(geom%w1)
      do i = geom%w1,geom%w2
        z = z_mid(i) - wtd
        pressure = atm * (1 + z/dist_atm)
        met_Cw = methaneP(i) / (geom%dzP(i) * por)! concentration in water
        co2_Cw = co2P(i) / (geom%dzP(i) * por)
        oxy_Cw = oxygenP(i) / (geom%dzP(i) * por)
        ppi(2) = met_Cw * Rgas * tempP(i) / met_alphaP(i) ! CH4 partial pressure
        ppi(3) = co2_Cw * Rgas * tempP(i) / co2_alphaP(i) ! CO2
        ppi(4) = oxy_Cw * Rgas * tempP(i) / oxy_alphaP(i) ! O2
        if (sum(ppi) > ssr * pressure) then
           fact = (sum(ppi) - pressure) / sum(ppi)
           do n = 1,4
              dppi(n) = fact * ppi(n) ! change in each gas [Pa]
           end do
           mrateP(i) = -k * dppi(2) / (Rgas * tempP(i)) * geom%dzP(i) * por
           crateP(i) = -k * dppi(3) / (Rgas * tempP(i)) * geom%dzP(i) * por
           orateP(i) = -k * dppi(4) / (Rgas * tempP(i)) * geom%dzP(i) * por
        end if
      end do

      if(geom%airQ) then
        mebu_tot = sum(mrateP)
        mrateP(geom%a2) = - mebu_tot ! bubbles reseased to deepest air layer
        cebu_tot = sum(crateP)
        crateP(geom%a2) = - cebu_tot
        oebu_tot = sum(orateP)
        orateP(geom%a2) = - oebu_tot
        ! ebu_out is already 0, no need to set again
      else
        mebu_out = sum(mrateP) ; mebu_tot = mebu_out
        cebu_out = sum(crateP) ; cebu_tot = cebu_out
        oebu_out = sum(orateP) ; oebu_tot = oebu_out
      end if

    end if
  end subroutine ebullition_rate

  pure function find_max_oxidation_dt(methaneP,met_alphaP,oxygenP,oxy_alphaP, &
    tempP,nz,geom,por) result(max_dt)

    use methane_parameters, only: Rgas,T0,V_oxid_max,delta_E_oxid,K_CH4,K_O2, &
                                  time_scale

    real(dp), intent(in)       :: methaneP(nz),met_alphaP(nz),oxygenP(nz), &
                                  oxy_alphaP(nz),tempP(nz),por
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    integer                    :: i
    real(dp)                   :: met_oxidP(nz),Cw_met,Cw_oxy,V_oxid_0, &
                                  V_oxidP(nz),turnover_rate_met, &
                                  turnover_rate_oxy,turnover_rate,derivative, &
                                  new_dt
    real(dp)                   :: max_dt

    max_dt = 24*3600 ! One day. Just to set something, before calculating actual
                     ! values.
    V_oxid_0 = V_oxid_max
    V_oxidP = V_oxid_0 * exp(delta_E_oxid / Rgas * (1/T0 - 1/tempP))

    met_oxidP = 0
    if (geom%airQ) then
      do i = geom%a1, geom%a2
        Cw_met = met_alphaP(i) * methaneP(i) / (por * geom%dzP(i))
        Cw_oxy = oxy_alphaP(i) * oxygenP(i) / (por * geom%dzP(i))
        met_oxidP(i) = V_oxidP(i) * (Cw_met / (K_CH4 + Cw_met)) * &
                                    (Cw_oxy / (K_O2  + Cw_oxy)) * geom%dzP(i)
      end do
    end if

    if (geom%waterQ) then
      do i = geom%w1, geom%w2
        Cw_met = methaneP(i) / (por * geom%dzP(i))
        Cw_oxy = oxygenP(i) / (por * geom%dzP(i))
        met_oxidP(i) = V_oxidP(i) * (Cw_met / (K_CH4 + Cw_met)) * &
             (Cw_oxy / (K_O2  + Cw_oxy)) * geom%dzP(i)
      end do
    end if

    do i = 1,nz
      if (met_oxidP(i) /= 0) then
        derivative = met_oxidP(i)
        turnover_rate_met = derivative / methaneP(i)
        turnover_rate_oxy = 2 * derivative / oxygenP(i)
        turnover_rate = max(turnover_rate_met,turnover_rate_oxy)
        new_dt = time_scale * 1 / turnover_rate
        if (new_dt < max_dt) then
          max_dt = new_dt
        end if
      end if
    end do
  end function find_max_oxidation_dt

  pure subroutine methane_oxidation(met_oxidP,met_oxid_tot, &
       co2_oxidP,co2_oxid_tot,oxy_oxidP,oxy_oxid_tot, &
       methaneP,met_alphaP,oxygenP,oxy_alphaP,tempP,nz,geom,por)

    use methane_parameters, only: Rgas,T0,V_oxid_max,delta_E_oxid,K_CH4,K_O2

    real(dp), intent(out)      :: met_oxidP(nz),met_oxid_tot, &
                                  co2_oxidP(nz),co2_oxid_tot,oxy_oxidP(nz),oxy_oxid_tot
    real(dp), intent(in)       :: methaneP(nz),met_alphaP(nz), &
                                  oxygenP(nz),oxy_alphaP(nz),tempP(nz),por
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    integer                    :: i
    real(dp)                   :: Cw_met,Cw_oxy,V_oxid_0,V_oxidP(nz)

    ! Dual-substrate Michaelis-Menten kinetics, Arah and Stephen (1998) Eq. 8.
    ! Assume a water phase that is in equilibrium with the gas phase,
    ! even for peat-air layers. No oxidation in the surface water above peat.

    oxy_oxidP = 0  ! initialize, try out if it matters...

    V_oxid_0 = V_oxid_max ! Eq. 13 without root distribution term
      ! Cannot use the root distribution term, because our epsilon_r(z) is
      ! non-zero at z=0, so the maximum of epsilon_r(z)/z would be infinity

    ! Eq. 10 in Arah & Stephen 1998
    V_oxidP = V_oxid_0 * exp(delta_E_oxid / Rgas * (1/T0 - 1/tempP))

    met_oxidP = 0
    if (geom%airQ) then
      do i = geom%a1, geom%a2
        Cw_met = met_alphaP(i) * methaneP(i) / (por * geom%dzP(i))
        Cw_oxy = oxy_alphaP(i) * oxygenP(i) / (por * geom%dzP(i))
        met_oxidP(i) = -V_oxidP(i) * (Cw_met / (K_CH4 + Cw_met)) * &
                                    (Cw_oxy / (K_O2  + Cw_oxy)) * geom%dzP(i)
        co2_oxidP(i) = -met_oxidP(i)
        oxy_oxidP(i) = 2._dp * met_oxidP(i)
      end do
    end if

    if (geom%waterQ) then
      do i = geom%w1, geom%w2
        Cw_met = methaneP(i) / (por * geom%dzP(i))
        Cw_oxy = oxygenP(i) / (por * geom%dzP(i))
        met_oxidP(i) = -V_oxidP(i) * (Cw_met / (K_CH4 + Cw_met)) * &
                                    (Cw_oxy / (K_O2  + Cw_oxy)) * geom%dzP(i)
        co2_oxidP(i) = -met_oxidP(i)
        oxy_oxidP(i) = 2._dp * met_oxidP(i)
      end do
    end if

    met_oxid_tot = sum(met_oxidP)
    co2_oxid_tot = sum(co2_oxidP)
    oxy_oxid_tot = sum(oxy_oxidP)
  end subroutine methane_oxidation

  pure function find_max_respiration_dt(nz,geom,oxygenP,tempP,alphaP,por) &
    result(max_dt)

    use methane_parameters, only: deltaE_R,Rgas,V_R0,K_R,T0,time_scale

    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: oxygenP(nz),tempP(nz),alphaP(nz),por
    real(dp)                   :: max_dt
    real(dp)                   :: respiP(nz),derivative,turnover_rate,new_dt
    integer                    :: i
    real(dp)                   :: Cw,V_R(nz)

    ! If dC/dt = -k*C, then k is the decay constant, or turover rate, and 1/k
    ! is turnover time. And a safe timestep for Runge-Kutta 4 is something less
    ! than the turnover time, depending how much precision you want. We choose
    ! 0.5. Here we calculate the rate of the chemical reaction used to model
    ! respiration, and then divide by oxygen content, so effectively getting
    ! just the k out of the k*C.
    !
    ! The constants related to respiration should not be repeated in this
    ! function, not good design. Will think of a more permanent design later.

    max_dt = 24*3600 ! One day. Just to set something, before calculating actual
                     ! values.
    V_R(:) = V_R0 * exp(deltaE_R / Rgas * (1/T0 - 1/tempP(:)))
    respiP = 0
    if (geom%airQ) then
      do i = geom%a1, geom%a2
        Cw = oxygenP(i) / (geom%dzP(i) * por) * alphaP(i)
        respiP(i) = V_R(i) * Cw / (K_R + Cw) * geom%dzP(i)
      end do
    end if

    if (geom%waterQ) then
      do i = geom%w1, geom%w2
        Cw = oxygenP(i) / (geom%dzP(i) * por)
        respiP(i) = V_R(i) * Cw / (K_R + Cw) * geom%dzP(i)
      end do
    end if

    do i = 1,nz
      if (respiP(i) /= 0) then
        derivative = respiP(i)
        turnover_rate = derivative / oxygenP(i)
        new_dt = time_scale * 1 / turnover_rate
        if (new_dt < max_dt) then
          max_dt = new_dt
        end if
      end if
    end do
  end function find_max_respiration_dt

  pure subroutine oxygen_respiration(oxy_respiP,oxy_respi_tot,co2_respiP,co2_respi_tot, &
       nz,geom,oxygenP,tempP,alphaP,por)

    use methane_parameters, only: deltaE_R,Rgas,V_R0,K_R,T0

    real(dp), intent(out)      :: oxy_respiP(nz),oxy_respi_tot,co2_respiP(nz),co2_respi_tot
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: oxygenP(nz),tempP(nz),alphaP(nz),por
    integer                    :: i
    real(dp)                   :: Cw,V_R(nz)

    ! This subroutine implements Arah and Stephen (1998) Eq. 7
    !
    ! respiP:    respiration rate in each gridbox [mol s-1]
    ! respi_tot: total respiration rate in each gridbox [mol s-1]
    ! Cw:        water phase oxygen concentration in this layer

    V_R(:) = V_R0 * exp(deltaE_R / Rgas * (1/T0 - 1/tempP(:))) ! A&S Eq. 10
    oxy_respiP = 0
    co2_respiP = 0
    if (geom%airQ) then
      do i = geom%a1, geom%a2
        Cw = oxygenP(i) / (geom%dzP(i) * por) * alphaP(i)
        oxy_respiP(i) = -V_R(i) * Cw / (K_R + Cw) * geom%dzP(i)
        co2_respiP(i) = - oxy_respiP(i) ! peat + O2 -> CO2 + H2O
      end do
    end if
    if (geom%waterQ) then
      do i = geom%w1, geom%w2
        Cw = oxygenP(i) / (geom%dzP(i) * por)
        oxy_respiP(i) = -V_R(i) * Cw / (K_R + Cw) * geom%dzP(i)
        co2_respiP(i) = - oxy_respiP(i) ! peat + O2 -> CO2 + H2O
      end do
    end if
    oxy_respi_tot = sum(oxy_respiP)
    co2_respi_tot = sum(co2_respiP)
  end subroutine oxygen_respiration

  pure subroutine root_transport(root_transportP,gas_flux,gasP,nz,geom, &
    conductanceP,alphaP,gas_Ca,por)
    real(dp), intent(out)      :: root_transportP(nz),gas_flux
    real(dp), intent(in)       :: gasP(nz),conductanceP(nz),alphaP(nz),gas_Ca, &
                                  por
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    integer                    :: i
    real(dp)                   :: cons ! gas gas-phase concentration in layer i
    ! gas_flux: flux from atmosphere

    ! If there is water above the peat, it does not contain roots, so no
    ! transport for surface water layer. (Alhough, we transport gas from roots
    ! to atmosphere, without accounting for the length of the plant "pipes"
    ! above peat surface. So assume plant stems are always above surface water,
    ! and that the aboveground part has no resistance.)

    ! In essence, root transport is diffusion in the air inside "plant tubes"
    ! (aerenchyma), so kind of funny to calculate root transport also for air
    ! layers because gases diffuse from air layers anyway. But peat tortuosity
    ! is assumed to be 10, and root tortuosity only 1.5, so even for air layers
    ! root transport is faster (but root cross section area is small).

    i = 1
    if (geom%surfaceQ .or. geom%dzP(i) == 0) then
      root_transportP(i) = 0
    else
      cons = gasP(i) / (por * geom%dzP(i))
      root_transportP(i) = conductanceP(i) * (gas_Ca - cons)
      ! This will always be an air layer. geom%w1 cannot be 1.
    end if

    do i = 2,nz
      if (geom%dzP(i) == 0) then
        root_transportP(i) = 0 
      else if (geom%waterQ .and. geom%w1 <= i) then
        cons = gasP(i) / (alphaP(i) * por * geom%dzP(i))
        root_transportP(i) = conductanceP(i) * (gas_Ca - cons)
      else
        cons = gasP(i) / (por * geom%dzP(i))
        root_transportP(i) = conductanceP(i) * (gas_Ca - cons)
      end if
    end do

    gas_flux = sum(root_transportP)
  end subroutine root_transport

  pure function find_max_exudation_dt(turnoverrate) result(max_dt)
    use methane_parameters, only: time_scale
    real(dp), intent(in) :: turnoverrate
    real(dp)             :: max_dt
    real(dp)             :: turnovertime
    turnovertime = 1 / turnoverrate
    max_dt = time_scale * turnovertime
  end function find_max_exudation_dt

  pure subroutine methane_from_peat(mrate_catoP,met_cato_tot,crate_catoP,co2_cato_tot, &
       nz,geom,tempP,fm)
    ! For now methane production from peat respiration is calculated only 
    ! for the catotelm. The acrotelm part is not included, regardless of 
    ! wheter oxic or anoxic conditions prevail. No actual catotelm is defined, 
    ! but the water table is used to determine an "instantaneous" acro and cato. 
    ! The acrotelm in this model is dry, therefore I would assume that the decay 
    ! would produce mostly CO2. The anaerobic respiration is modeled using 
    ! a Q10 model, see Schuldt et al. Biogeosciences, Modelling Holocene 
    ! carbon accumulation and methane emissions of boreal wetlands, 2013.
    !
    ! The model described in Schuldt et al. includes the decomposition in the 
    ! catotelm and the acrotelm (both the anoxic and the oxic part). The decay 
    ! rates are then summed and the production of methane is calculated using 
    ! a single factor for the ratio of CH4 and CO2. This seems quite crude, as 
    ! the turnover times are very different (oxic acro: 15, anoxic acro: 43 
    ! and cato: 30000 years). One would expect that the processes are different
    ! and the ratio of CH4 and CO2 would vary considerably between the different 
    ! parts of the peat column. According to Schuldt et al. this factor is obtained 
    ! from Scanlon and Moore (2000), however that paper does not deal with methane. 
    ! The paper also says ... scaled by the ratio of CO2 to CH4 emissions (fm=0.4). 
    ! This would mean that CO2/CH4 = 0.4. They multiply the carbon decay rate by fm 
    ! to get the methane production, which is incorrect if CO2/CH4=0.4. 
    ! Anyway I used the same fraction as Schuldt et al., assuming that each mol of C
    ! respired in the catotelm produces 0.4 mol CH4 on average.
    !
    ! Furthermore, I had a look at the values given in mo_cbal_cpols.f90
    ! tau_Cpool_acro_aboveWT = 14.92_dp*365._dp
    ! tau_Cpool_acro_belowWT = 42.64_dp*365._dp
    ! tau_Cpool_cato         = 11111._dp*365._dp
    ! rho_Cpool_acro         = 4579.14_dp
    ! rho_Cpool_cato         = 6277.73_dp
    ! In Schuldt et al. tau_Cpool_cato = 30000 years. The comments indicate that they 
    ! would consider acro_aboveWT as a source of only CO2 (not in line with the paper);
    ! acro_belowWT and cato would also be sources of methane. This is in line with what 
    ! I have assumed. However, as we assume everyting below the WT is catotelm we get
    ! very low turnover rates. Should we just adjust the turnover time, or should we 
    ! somehow add an anoxic acrotelm?
    !
    ! I am not able to get the same carbon pool densities as Schuldt et al. I get 
    ! 3939.72 mol(C) m-3 while Schuldt et al. use 6277.73 mol(C) m-3 for the cato,
    ! for now I trust my values more. Real world values might be something else.
    !
    ! How would the water stress factor influence the results, might be important
    ! in the acrotelm?
    !
    use methane_parameters, only: alpha,Q10,Tref,tau_C_cato,spy,rho_cato,f_C_cato,M_C
    integer,  intent(in)       :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: tempP(nz),fm
    real(dp), intent(out)      :: mrate_catoP(nz),met_cato_tot,crate_catoP(nz),co2_cato_tot
    real(dp) :: k_catoP(nz),tau_cato,rho_C_cato
    integer  :: i

    ! Turn over time of catotelm carbon pool, years to seconds
    tau_cato = tau_C_cato*spy ! 9.4670778e11_dp
    ! Carbon molar density of catotelm, mol(C) m-3
    rho_C_cato = f_C_cato*rho_cato/M_C ! 3939.72, differs from Schuldt et al. 6277.73
    k_catoP = 0
    mrate_catoP = 0
    crate_catoP = 0
    ! for each layer below WT (per m2)
    if(geom%waterQ) then
      do i = geom%w1, geom%w2
        ! eq(13) in Schuldt et al. 2013. [s-1]
        k_catoP(i) = alpha * Q10**((tempP(i)-Tref)/10._dp) / tau_cato
        ! eq(14) in Schuldt et al. 2013. [mol(CH4) s-1]
        mrate_catoP(i) = k_catoP(i)*rho_C_cato*geom%dzP(i)*fm
        crate_catoP(i) = k_catoP(i)*rho_C_cato*geom%dzP(i)*(1._dp-fm)
      end do
    end if
    met_cato_tot = sum(mrate_catoP)
    co2_cato_tot = sum(crate_catoP)
  end subroutine methane_from_peat

  subroutine methane_from_exu(met_rateP,met_exu_tot, &
       co2_rateP,co2_exu_tot,oxy_rateP,oxy_exu_tot,nz,geom, &
       rootsP,oxygenP,exu_decayrate,f_methane)
    ! The total input of C from root exudates pool is 'exu_decayrate'. This is
    ! distributed according to the root distribution. For the part of exudates
    ! below WTD, 'f_methane' fraction is converted to methane.
    ! (Later when adding CO2, the rest is converted to CO2.)

    use methane_parameters, only: ny
    real(dp), intent(out)      :: met_rateP(nz),met_exu_tot, &
                                  co2_rateP(nz),co2_exu_tot, &
                                  oxy_rateP(nz),oxy_exu_tot
    integer,  intent(in)       :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: rootsP(nz),oxygenP(nz),exu_decayrate,f_methane
    real(dp)                   :: max_rateP(nz)
    integer  :: i

    met_rateP = 0
    co2_rateP = 0
    oxy_rateP = 0

    ! above WT: exu + O2 -> CO2 + H2O
    ! all exu is oxidized above WT
    if (geom%airQ) then
       do i = geom%a1, geom%a2
          co2_rateP(i) = rootsP(i) * exu_decayrate
          oxy_rateP(i) = - co2_rateP(i)
       end do
    end if

    ! below WT: exu -> CH4 + CO2
    ! and oxygen inhibition: exu + O2 -> CO2 + H2O
    if (geom%waterQ) then
      do i = geom%w1, geom%w2
        max_rateP(i) = rootsP(i) * exu_decayrate * f_methane
        met_rateP(i) = max_rateP(i) / (1._dp + ny*oxygenP(i))
        co2_rateP(i) = (max_rateP(i) - met_rateP(i)) &
             + rootsP(i) * exu_decayrate * (1._dp-f_methane)
        oxy_rateP(i) = - (max_rateP(i) - met_rateP(i))
      end do
    end if

    met_exu_tot = sum(met_rateP)
    co2_exu_tot = sum(co2_rateP)
    oxy_exu_tot = sum(oxy_rateP)

  end subroutine methane_from_exu

  pure subroutine diffusion(dgasdtP,flux,gasP,nz,geom,C_air,DcondP,alphaP,invpordzP)
    real(dp), intent(out)      :: dgasdtP(nz),flux
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: gasP(nz),C_air,DcondP(nz),alphaP(nz),invpordzP(nz)
    real(dp)                   :: d(nz+1), X(nz+1) ! helper variables
    integer                    :: h
    
    ! flux: flux from atmosphere

    ! To be precise, we should not calculate diffusion flux based only on gas
    ! concentration, but also take into account the change solubility depending
    ! on hydrostatic pressure, and temperature. That is, to use chemical
    ! potentials instead of concentrations. But these effects should be small,
    ! under 5%, so for now they are not included.
     
    d = 0; X = 0; dgasdtP = 0; h = 0
    if (geom%airQ) h = geom%w1-geom%a2-1 ! 1 if we have a dz=0-layer, 0 otherwise
    
    if (geom%surfaceQ) then
       d(1) = alphaP(1)*C_air
       d(2) = gasP(1)/geom%dzP(1)
       d(3:) = gasP(2:)*invpordzP(2:)
       X(1:nz) = DcondP(1:nz)*(d(1:nz) - d(2:nz+1))
       dgasdtP(1:nz) = X(1:nz) - X(2:nz+1)
       flux = DcondP(1)*(d(1) - d(2))
    else if (geom%airQ) then
       d(1) = C_air
       d(2:geom%a2+1) = gasP(1:geom%a2)*invpordzP(1:geom%a2)
       d(geom%w1+1-h:geom%w2+1-h) = gasP(geom%w1:geom%w2)*invpordzP(geom%w1:geom%w2)
       ! Construct the X-vector
       if (.not. geom%airq .or. geom%w1 - geom%a2 == 1) then
          X(1:nz) = DcondP(1:nz)*(d(1:nz) - d(2:nz+1))
          ! Air-water interface correction:
          X(geom%a2+1) = X(geom%a2+1) - DcondP(geom%w1)*d(geom%w1)*(1 - alphaP(geom%a2))
          dgasdtP(1:nz-1) = X(1:nz-1) - X(2:nz)
       else
          X(1:geom%a2) = DcondP(1:geom%a2)*(d(1:geom%a2) - d(2:geom%a2+1))
          X(geom%a2+1:nz-(geom%w1-geom%a2-1)) = DcondP(geom%w1:nz)*(d(geom%a2+1:nz-1) - d(geom%a2+2:nz))
          ! Air-water interface correction:
          X(geom%a2+1) = X(geom%a2+1) - DcondP(geom%w1)*d(geom%w1-1)*(1 - alphaP(geom%a2)) 
          dgasdtP(1:geom%a2) = X(1:geom%a2) - X(2:geom%a2+1)
          dgasdtP(geom%w1:nz) = X(geom%a2+1:nz-h) - X(geom%a2+2:nz-h+1)
       end if
       flux = DcondP(1)*(d(1) - d(2))
    else
       d(2) = alphaP(2)*C_air
       d(3:) = gasP(2:)*invpordzP(2:)
       X(1:nz) = DcondP(1:nz)*(d(1:nz) - d(2:nz+1))
       dgasdtP(1:nz) = X(1:nz) - X(2:nz+1)
       dgasdtP(1) = 0
       flux = DcondP(2)*(d(2) - d(3))
    end if
    dgasdtP(nz) = X(nz-h)
    
  end subroutine diffusion

  pure function find_max_diffusion_dt(nz,geom,DcondP) result(max_dt)
    ! For time stepping stability, apply Von Neumann stability condition:
    ! dt <= 1/2 * dx/cond, where dx and centerpoint-to-centerpoit donductances
    ! are arrays, and the smallest ratio needs to be found.

    use methane_parameters, only: time_scale

    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: DcondP(nz)
    real(dp)                   :: max_dt
    integer                    :: i
    real(dp)                   :: valP(nz)
    ! DcondP: array (nz elements) of diffusion conductances between boxes in
    ! the peat gas profile. DcondP(1) is the conductance from box 1 to
    ! atmosphere. DconP(i) is the conductance between boxes i-1 and i. In case
    ! of geom%dz(i) = 0 for layer i, both DcondP(i-1) and Dcond(i) are assumed
    ! to give the same conductance from layer i+1 to layer i+1.

    do i = 1,nz-1
      if (geom%dzP(i) == 0) then
        valP(i) = huge(1.0_dp)
      else
        valP(i) = 1 / (max(DcondP(i),DcondP(i+1)) / geom%dzP(i))
      end if
    end do
    if (geom%dzP(nz) == 0) then
      valP(nz) = huge(1.0_dp)
    else
      valP(nz) = 1 / (DcondP(nz) / geom%dzP(nz))
    end if

    max_dt = time_scale * minval(valP)
  end function find_max_diffusion_dt

  pure function diffusion_conductances(nz,geom,DaP,Dwp,alphaP) &
                result(DcondP)
    ! prepare diffusion conductances between box centerpoints. DcondP(1) for
    ! the dzP(1)/2 distance from the top box center to atmosphere.
    use methane_parameters, only: peat_air_coeff
    
    real(dp)                   :: DcondP(nz)
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: DaP(nz),DwP(nz),alphaP(nz)
    integer                    :: i,i1,i2
    real(dp)                   :: r1,r2

    ! The model speed depends a lot on the time step. As diffusion in
    ! air is around 1000 times faster than diffusion in water, we are
    ! constrained by the thickness of the thinnest of the air
    ! layers. Since diffusion is at any rate enough fast to get the
    ! gases out on the right day (normal model resolution), we can
    ! just as well pretend that the conductance of the air layers are
    ! as if their thicknesses were at least 10 cm. The parameter
    ! fakemindz does exactly that. Speeds up model by a factor of 3
    ! and only very minimally affects output fluxes.
    real, parameter :: fakemindz=0.1

    ! The conductance over air-water interface, when used above in 'diffusion',
    ! assumes concentrations in water are used in the calculation then.

    if (geom%surfaceQ) then
      DcondP(1) = 1./peat_air_coeff * DwP(1) / (0.5 * max(fakemindz, geom%dzP(1)))
      ! because above-surface-water has no peat
      r1 = (0.5*geom%dzP(1)) / DwP(1) / peat_air_coeff
      r2 = (0.5*geom%dzP(2)) / DwP(1)
      DcondP(2) = 1/(r1 + r2)
      do i = geom%w1 + 1, geom%w2
        r1 = (0.5*geom%dzP(i-1)) / DwP(i-1)
        r2 = (0.5*geom%dzP(i)) / DwP(i)
        DcondP(i) = 1/(r1 + r2)
      end do
    else if (geom%airQ) then
      DcondP(1) = DaP(1) / (0.5 * max(fakemindz, geom%dzP(1)))
      do i = geom%a1 + 1, geom%a2
        r1 = (0.5*max(fakemindz, geom%dzP(i-1))) / DaP(i-1)
        r2 = (0.5*max(fakemindz, geom%dzP(i))) / DaP(i)
        DcondP(i) = 1/(r1 + r2)
      end do
      if (geom%waterQ) then
        i1 = geom%a2
        i2 = geom%w1
        ! this formula would benefit from documentation and derivation
        DcondP(i1+1) = DaP(i1) * DwP(i2) / &
          ( 0.5*max(fakemindz, geom%dzP(i2))*DaP(i1) + 0.5*geom%dzP(i1))*DwP(i2)*alphaP(i2)
        ! alphaP(i2) or alphaP(i1) can be debated, but let's take alphaP(i2)
        if (i1 + 1 < i2) then
          DcondP(i2) = DcondP(i1+1)
          ! just so later we don't need to care which one, as they are same
        end if
        do i = geom%w1 + 1, geom%w2
          r1 = (0.5*geom%dzP(i-1)) / DwP(i-1)
          r2 = (0.5*geom%dzP(i)) / DwP(i)
          DcondP(i) = 1/(r1 + r2)
        end do
      end if
    else ! so water is at peat surface level
      DcondP(1) = 0 ! dzP(1) must be 0, so this should never be used
      i = geom%w1
      DcondP(i) = DwP(i) / (0.5 * geom%dzP(i))
      do i = geom%w1 + 1, geom%w2
        r1 = (0.5*geom%dzP(i-1)) / DwP(i-1)
        r2 = (0.5*geom%dzP(i)) / DwP(i)
        DcondP(i) = 1/(r1 + r2)
      end do
    end if
  end function diffusion_conductances

  pure function root_conductances(nz,geom,DaP,rootsP,root_km,root_mass) &
                result(conductanceP)
    use methane_parameters, only: root_tortuosity, peat_air_coeff
    integer, intent(in)        :: nz
    type(geometry), intent(in) :: geom
    real(dp), intent(in)       :: DaP(nz),rootsP(nz),root_km,root_mass
    real(dp)                   :: conductanceP(nz)
    real(dp)                   :: resistanceP(nz)
    real(dp)                   :: DaP_nopeat(nz)
    integer                    :: i

    ! DaP: air diffusion coefficients
    ! resistanceP: box-to-box resistances

    ! No peat inside stems, but input argument DaP contains their
    ! contribution, so it has to be negated here:
    DaP_nopeat = DaP/peat_air_coeff

    ! - We assume that surface water does not resist aerenchyma transport
    ! - Special care to deal with the possible dz=0 layer

    i = 1
    if (geom%surfaceQ .or. geom%dzP(i) == 0) then
      resistanceP(i) = 0
    else
      resistanceP(i) = 0.5*geom%dzP(i)/DaP(i)
    end if
    do i = 2,nz
      if (geom%dzP(i) == 0) then
        resistanceP(i) = 0.5*geom%dzP(i-1)/DaP(i-1)
      else
        if (resistanceP(i-1) == 0) then
          resistanceP(i) = 0.5*geom%dzP(i)/DaP(i)
        else
          resistanceP(i) = 0.5*geom%dzP(i)/DaP(i) + 0.5*geom%dzP(i-1)/DaP(i-1)
        end if
      end if
    end do

    i = 1
    if (resistanceP(i) == 0) then
      conductanceP(i) = 0 ! should never be used, just have put something
    else
      conductanceP(i) = 1.0 / resistanceP(i)
    end if
    do i = 2,nz
      conductanceP(i) = 1.0 / sum(resistanceP(1:i))
    end do

    conductanceP = conductanceP * rootsP * root_km * root_mass / root_tortuosity
  end function root_conductances

  pure elemental function methane_D_air(temperature) result(Da)
    ! methane diffusion coefficient for medium consisting of xx% peat,
    ! yy% air, depending on temperature
    use methane_parameters, only: peat_air_coeff
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: Da
    real(dp), parameter  :: Da_273 = 1.9e-5_dp ! [m2 s-1]
                            ! did not yet check this.
    ! Arah and Stephen (1998), Eq. 11
    Da = peat_air_coeff * Da_273 * (temperature/273)**1.82
  end function methane_D_air

  pure elemental function methane_D_water(temperature) result(Dw)
    ! methane diffusion coefficient for medium consisting of xx% peat,
    ! yy% water, depending on temperature
    use methane_parameters, only: peat_water_coeff
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: Dw
    real(dp), parameter  :: Dw_298 = 1.5e-9_dp ! [m2 s-1]

    ! Arah and Stephen (1998), Eq. 11
    Dw = peat_water_coeff * Dw_298 * (temperature/298)
    ! if (temperature < 273.15) Dw = 0
  end function methane_D_water

  pure elemental function methane_solubility(temperature) result(alpha)
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: alpha ! [mol(CH4,water) m(water)-3 mol(CH4,air)-1 m(air)3]
    real(dp), parameter  :: a = 1.3e-3_dp, b = 1700._dp, c = 298.0_dp
    real(dp), parameter  :: Rgas = 8.3144621_dp ! [J mol-1 K-1]
    real(dp) :: Hcp ! [mol(CH4) m(H2O)-3 Pa-1]
    ! Tang et al. 2010, eq. A4
    ! see also http://www.henrys-law.org/henry-3.0.pdf
    Hcp = a * exp(b * (1._dp/temperature - 1._dp/c) ) * 9.86923266716e-3_dp
    alpha = Rgas * temperature * Hcp
  end function methane_solubility

  pure elemental function co2_D_air(temperature) result(Da)
    ! CO2 diffusion coefficient for medium consisting of xx% peat,
    ! yy% air, depending on temperature
    use methane_parameters, only: peat_air_coeff
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: Da
    real(dp), parameter  :: Da_273 = 1.47e-5_dp ! [m2 s-1]
                            ! from IIyama and Hasegawa (2005), Fig. 2

    ! Tang et al. 2010, eq. A8
    Da = peat_air_coeff * Da_273 * (temperature/273.15_dp)**1.792_dp
  end function co2_D_air

  pure elemental function co2_D_water(temperature) result(Dw)
    ! CO2 diffusion coefficient for medium consisting of xx% peat,
    ! yy% water, depending on temperature
    use methane_parameters, only: peat_water_coeff
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: Dw
    real(dp), parameter  :: Dw_298 = 1.81e-6_dp ! [m2 s-1]

    ! Tang et al. 2010, eq. A12
    Dw = peat_water_coeff * Dw_298 * exp(-2032.6_dp/temperature)
  end function co2_D_water

  pure elemental function co2_solubility(temperature) result(alpha)
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: alpha ! [mol(CO2,water) mol(CO2,air)]
    real(dp), parameter  :: a = 3.4e-2_dp, b = 2400._dp, c = 298.0_dp
    real(dp), parameter  :: Rgas = 8.3144621_dp ! [J mol-1 K-1]
    real(dp) :: Hcp ! [mol(CH4) m(H2O)-3 Pa-1]
    ! Tang et al. 2010, eq. A3
    ! see also http://www.henrys-law.org/henry-3.0.pdf
    Hcp = a * exp(b * (1._dp/temperature - 1._dp/c) ) * 9.86923266716e-3_dp
    alpha = Rgas * temperature * Hcp
  end function co2_solubility

  pure elemental function oxygen_D_air(temperature) result(Da)
    ! oxygen diffusion coefficient for medium consisting of xx% peat,
    ! yy% air, depending on temperature
    use methane_parameters, only: peat_air_coeff
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: Da
    real(dp), parameter  :: Da_273 = 1.8e-5_dp ! [m2 s-1]
                            ! from IIyama and Hasegawa (2005), Fig. 2

    ! Arah and Stephen (1998), Eq. 11
    Da = peat_air_coeff * Da_273 * (temperature/273._dp)**1.82_dp
  end function oxygen_D_air

  pure elemental function oxygen_D_water(temperature) result(Dw)
    ! methane diffusion coefficient for medium consisting of xx% peat,
    ! yy% water, depending on temperature
    use methane_parameters, only: peat_water_coeff
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: Dw
    real(dp), parameter  :: Dw_298 = 2.4e-9_dp ! [m2 s-1]

    ! Arah and Stephen (1998), Eq. 11
    Dw = peat_water_coeff * Dw_298 * (temperature/298._dp)
    ! if (temperature < 273.15) Dw = 0
  end function oxygen_D_water

  pure elemental function oxygen_solubility(temperature) result(alpha)
    real(dp), intent(in) :: temperature ! [K]
    real(dp)             :: alpha ! [mol(CO2,water) mol(CO2,air)]
    real(dp), parameter  :: a = 1.3e-3_dp, b = 1500._dp, c = 298.0_dp
    real(dp), parameter  :: Rgas = 8.3144621_dp ! [J mol-1 K-1]
    real(dp) :: Hcp ! [mol(O2) m(H2O)-3 Pa-1]
    ! Tang et al. 2010, eq. A2
    ! see also http://www.henrys-law.org/henry-3.0.pdf
    Hcp = a * exp(b * (1._dp/temperature - 1._dp/c) ) * 9.86923266716e-3_dp
    alpha = Rgas * temperature * Hcp
  end function oxygen_solubility

  pure function calculate_root_fraction_alternative(lambda_root,nz,levelsP,peat_surface) &
                result(rootsP)
    real(dp), intent(in)  :: lambda_root,levelsP(nz+1),peat_surface
    integer,  intent(in)  :: nz
    real(dp)              :: rootsP(nz),top,bottom
    integer               :: i,i_start
    ! roots are distributed according to:
    !     root_fraction(z) = C exp(-z/lambda_root)
    ! where C is a constant to normalize the integral of the distribution to 1.
    ! The integral from a to b of exp(-z/lambda_root) is
    !     (exp(-a/lambda_root) - exp(-b/lambda_root)) * lambda_root
    ! and then we just normalize numerically.
    !
    ! The z's are positive numbers here, even though we are underground.

    i_start = 1
    if (levelsP(1) < peat_surface) then ! layer is is water above surface
      rootsP(1) = 0
      i_start = 2
    end if

    do i = i_start,nz
      top    = levelsP(i)   - peat_surface
      bottom = levelsP(i+1) - peat_surface
      rootsP(i) = &
       430*(exp(-(top-.105)**2/lambda_root**2) + exp(-(bottom-.105)**2/lambda_root**2))/2 + 6
    end do
    rootsP = rootsP / sum(rootsP)
  end function calculate_root_fraction_alternative

end program methane_main
