module methane_parameters
  implicit none

  !integer, parameter :: dp = selected_real_kind(15,300) !64bit double precision
  integer, parameter :: dp = kind(1.) ! float, ~4% speedup but no meaningful loss of precision

  real(dp), parameter :: min_dz = 0.01_dp ! meters
  ! minimum thickness for boxes, because small boxes require a smaller time step

  integer, parameter :: n_flux_parts = 22
  ! indices for the bookkeeping arrays where we collect the output fluxes etc.
! *** methane transport ***
  !  1 = methane diffusion
  !  2 = methane plant transport
  !  3 = methane ebullition into air
  !  4 = methane ebullition total (when WTD is below peat surface, ebullition is
  !      released in the lowest air layer inside peat)
! *** methane loss & production ***
  !  5 = production of methane from exu (also not a flux)
  !  6 = production of methane from peat decay in catotelm (below WT)
  !  7 = oxidation of methane (not a flux out, just a sink)
! *** CO2 transport ***
  !  8 = CO2 diffusion
  !  9 = CO2 plant transport
  ! 10 = CO2 ebullition into air (CO2 ebullition is assumed to occur with methane)
  ! 11 = CO2 ebullition total (when WTD is below peat surface, ebullition is
  !      released in the lowest air layer inside peat)
! *** CO2 production ***
  ! 12 = production of CO2 from exu (also not a flux)
  ! 13 = production of CO2 from peat decay (also not a flux)
  ! 14 = production of CO2 from methane oxidation (not a flux out)
  ! 15 = production of CO2 from respiration (also not a flux)
! *** Oxygen transport ***
  ! 16 = oxygen diffusion
  ! 17 = oxygen plant transport
  ! 18 = oxygen ebullition into air (oxygen ebullition is assumed to occur with methane)
  ! 19 = oxygen ebullition total (when WTD is below peat surface, ebullition is
  !      released in the lowest air layer inside peat)
! *** Oxygen consumption ***
  ! 20 = oxygen consumed in oxidation of root exudates
  ! 21 = oxygen consumed in respiration
  ! 22 = oxygen consumed in oxidation of methane

  integer, parameter :: mdiff   =  1, &
                        mplant  =  2, &
                        mebuair =  3, &
                        mebutot =  4, &
                        mprod   =  5, &
                        mdecay  =  6, &
                        moxi    =  7, &
                        cdiff   =  8, &
                        cplant  =  9, &
                        cebuair = 10, &
                        cebutot = 11, &
                        cprod   = 12, &
                        cdecay  = 13, &
                        cresp   = 14, &
                        coxi    = 15, &
                        odiff   = 16, &
                        oplant  = 17, &
                        oebuair = 18, &
                        oebutot = 19, &
                        oprod   = 20, &
                        oresp   = 21, &
                        ooxi    = 22

  ! Scaling to choose a safe time step for numerics. Anything up to 1 should be
  ! safe, but values close to 1 might give oscillations (but only dampening
  ! type) in the solution of diffusion. 0.5 should be very safe. NO IT IS NOT!
  ! real(dp), parameter :: time_scale = 0.5

  ! Gas constant
  real(dp), parameter :: Rgas = 8.3144621_dp ! [J mol-1 K-1]

  ! Reference temperature for respiration and oxidation
  real(dp), parameter :: T0 = 283._dp ! [K]

  ! Atmospheric pressure [Pa]
  real(dp), parameter :: atm = 101325._dp ! atm pressure [Pa]

  ! ebu_hl: Ebullition half-life of supersaturated dissolved methane
  ! super saturation requirement for future use, now set to 1
  real(dp), parameter :: ssr = 1.0_dp

  ! methane from exu, oxygen inhibition on methanogenesis
  real(dp), parameter :: ny = 400._dp ! [m3 mol-1], Tang et al. 2010

  ! Oxygen respiration parameters:
  ! deltaE_R = 50000    ! [J mol-1]
  ! V_R0 = 4e-5_dp      ! [mol m-3(peat) s-1]
  ! K_R  = 0.22         ! [mol m-3(water)]

  ! Methane oxidation parameters
  ! V_oxid_max = 6.e-4_dp    ! [mol m-3(peat) s-1]
  ! delta_E_oxid = 50000.    ! [J mol-1]
  ! K_CH4 = 0.44             ! [mol m-3(water)]
  ! K_O2 = 0.33              ! [mol m-3(water)]

  ! peat_air_coeff = 0.1 (default) from IIyama and Hasegawa (2005), Fig. 2

  real(dp) :: por, peat_air_coeff, peat_water_coeff, &
       lambda_root, V_R0, K_R, K_CH4, K_O2, root_tortuosity, &
       root_km, f_exu, exu_pool_turnovertime, &
       deltaE_R, V_oxid_max, delta_E_oxid, ebu_hl, Q10, tau_C_cato, &
       time_scale, f_methane

  ! Methane from peat decay
  ! Mean water stress factor (beteween 0 and 1=no stress) not really a fixed parameter 
  real(dp), parameter  :: alpha = 1._dp ! set to 1 (I assume this is OK below the WT)
  ! Base for temperature dependent respiration
  ! real(dp), parameter  :: Q10 = 1.8_dp
  ! Q10 reference temperature, K
  real(dp), parameter  :: Tref = 273.15_dp
  ! Turn over time of catotelm carbon pool, 30000 years
  ! real(dp), parameter  :: tau_C_cato = 11111._dp
  ! seconds per year
  real(dp), parameter  :: spy = 31556926._dp
  ! Peat density of catotelm, g(peat) m-3 [Turunen et al. (2002)]
  real(dp), parameter  :: rho_cato = 9.1e4_dp ! 91 kg(peat) m-3
  ! Carbon fraction in catotelm peat, g(C) g(peat)-1 [Malmer and Wallen (2004)]
  ! Was not able to verify the unit. Probably mass ratio, not molar.
  real(dp), parameter  :: f_C_cato = 0.52_dp
  ! molar mass of Carbon, g(C) mol(C)-1 
  real(dp), parameter  :: M_C = 12.011_dp
  ! mol methane produced per mol carbon, mol(CH4) mol(C)-1
  real(dp), parameter  :: fm = 0.4_dp

contains

  subroutine get_parameters
      namelist /mcmc/ por, peat_air_coeff, peat_water_coeff, lambda_root, &
           V_R0, K_R, K_CH4, K_O2, root_tortuosity, root_km, f_exu, f_methane, &
           exu_pool_turnovertime, deltaE_R, V_oxid_max, delta_E_oxid, &
           ebu_hl, Q10, tau_C_cato, time_scale

      open(4, file='mcmc.nml')
      read(4, nml=mcmc)
      close(4)
  end subroutine get_parameters

end module methane_parameters
